searchData={"items":[{"type":"module","title":"segmented_cache","doc":"`segmented_cache` is a key/value pairs cache library implemented in rotating segments.\n\nFor more information, see the README, and the function documentation.","ref":"segmented_cache.html"},{"type":"function","title":"segmented_cache.delete_entry/2","doc":"Delete an entry in all ets segments.\n\nMight raise a telemetry error if the request fails:\n- name: `[segmented_cache, Name, delete_error]`\n- measurements: `#{}`\n- metadata: `t:delete_error/1`","ref":"segmented_cache.html#delete_entry/2"},{"type":"type","title":"segmented_cache.delete_error/1","doc":"Telemetry metadata with deletion error information.","ref":"segmented_cache.html#t:delete_error/1"},{"type":"function","title":"segmented_cache.delete_pattern/2","doc":"Delete a pattern in all ets segments.\n\nMight raise a telemetry error if the request fails:\n- name: `[segmented_cache, Name, delete_error]`\n- measurements: `#{}`\n- metadata: `t:delete_error/1`","ref":"segmented_cache.html#delete_pattern/2"},{"type":"type","title":"segmented_cache.entries_limit/0","doc":"Maximum number of entries per segment. When filled, rotation is ensued.","ref":"segmented_cache.html#t:entries_limit/0"},{"type":"function","title":"segmented_cache.get_entry/2","doc":"Get the entry for Key in cache.\n\nRaises telemetry span:\n- name: `[segmented_cache, Name, request, _]`\n- start metadata: `#{name => atom()}`\n- stop metadata: `t:hit/0`","ref":"segmented_cache.html#get_entry/2"},{"type":"type","title":"segmented_cache.hit/0","doc":"Telemetry metadata with cache hit information.","ref":"segmented_cache.html#t:hit/0"},{"type":"function","title":"segmented_cache.is_member/2","doc":"Check if Key is cached.\n\nRaises a telemetry span:\n- name: `[segmented_cache, Name, request, _]`\n- start metadata: `#{name => atom()}`\n- stop metadata: `t:hit/0`","ref":"segmented_cache.html#is_member/2"},{"type":"type","title":"segmented_cache.key/0","doc":"Dynamic type of _keys_ from cache clients.","ref":"segmented_cache.html#t:key/0"},{"type":"function","title":"segmented_cache.merge_entry/3","doc":"Merge a new entry into an existing one, or add it at the front if none is found.\n\nRace conditions considerations:\n- Two writers: `compare_and_swap` will ensure they both succeed sequentially\n- Any writers and the cleaner: under fifo, the writer modifies the record in place\n     and doesn't need to be concerned with rotation. Under lru, the same considerations\n     than for a `put_entry_front` apply.","ref":"segmented_cache.html#merge_entry/3"},{"type":"type","title":"segmented_cache.merger_fun/1","doc":"Merging function to use for resolving conflicts","ref":"segmented_cache.html#t:merger_fun/1"},{"type":"type","title":"segmented_cache.name/0","doc":"Cache unique name.","ref":"segmented_cache.html#t:name/0"},{"type":"type","title":"segmented_cache.opts/0","doc":"Configuration values for the cache.","ref":"segmented_cache.html#t:opts/0"},{"type":"function","title":"segmented_cache.put_entry/3","doc":"Add an entry to the first table in the segments.","ref":"segmented_cache.html#put_entry/3"},{"type":"function","title":"Possible race conditions: - segmented_cache.put_entry/3","doc":"- Two writers: another process might attempt to put a record at the same time. It this case,\n     both writers will attempt `ets:insert_new`, resulting in only one of them succeeding.\n     The one that fails, will retry three times a `compare_and_swap`, attempting to merge the\n     values and ensuring no data is lost.\n- One worker and the cleaner: there's a chance that by the time we insert in the ets table,\n     this table is not the first anymore because the cleaner has taken action and pushed it\n     behind.\n- Two writers and the cleaner: a mix of the previous, it can happen that two writers can\n     attempt to put a record at the same time, but exactly in-between, the cleaner rotates the\n     tables, resulting in the first writter inserting in the table that immediately becomes the\n     second, and the latter writter inserting in the recently treated as first, shadowing the\n     previous.\n\nTo treat the data race with the cleaner, after a successful insert,\nwe re-check the index, and if it has changed, we restart the whole operation again:\nwe can be sure that no more rotations will be triggered in a while,\nso the second round will be final.","ref":"segmented_cache.html#put_entry/3-possible-race-conditions"},{"type":"function","title":"Strategy considerations: - segmented_cache.put_entry/3","doc":"Under a fifo strategy, no other writes can happen, but under a lru strategy,\nmany other workers might attemp to move a record forward. In this case,\nthe forwarding movement doesn't modify the record, and therefore the `compare_and_swap`\noperation should succeed at once; then, once the record is in the front,\nall other workers shouldn't be attempting to move it.","ref":"segmented_cache.html#put_entry/3-strategy-considerations"},{"type":"type","title":"segmented_cache.scope/0","doc":"`m:pg` scope for cache coordination across distribution.","ref":"segmented_cache.html#t:scope/0"},{"type":"function","title":"segmented_cache.start/1","doc":"See `start_link/2` for more details","ref":"segmented_cache.html#start/1"},{"type":"function","title":"segmented_cache.start/2","doc":"See `start_link/2` for more details","ref":"segmented_cache.html#start/2"},{"type":"function","title":"segmented_cache.start_link/1","doc":"See `start_link/2` for more details","ref":"segmented_cache.html#start_link/1"},{"type":"function","title":"segmented_cache.start_link/2","doc":"Start and link a cache entity in the local node.\n\n`Name` must be an atom. Then the cache will be identified by the pair `{segmented_cache, Name}`,\nand an entry in persistent_term will be created and the worker will join a pg group of\nthe same name.\n`Opts` is a map containing the configuration.\n- `scope` is a `pg` scope. Defaults to `pg`.\n- `strategy` can be fifo or lru. Default is `fifo`.\n- `segment_num` is the number of segments for the cache. Default is `3`\n- `ttl` is the live, in minutes, of _each_ segment. Default is `480`, i.e., 8 hours.\n- `merger_fun` is a function that, given a conflict,\n    takes in order the old and new values and applies a merging strategy.\n    See the `t:merger_fun/1` type.","ref":"segmented_cache.html#start_link/2"},{"type":"type","title":"segmented_cache.strategy/0","doc":"Strategy for cache eviction.","ref":"segmented_cache.html#t:strategy/0"},{"type":"type","title":"segmented_cache.value/0","doc":"Dynamic type of _values_ from cache clients.","ref":"segmented_cache.html#t:value/0"},{"type":"extras","title":"Overview","doc":"# `segmented_cache`\n\n[![Actions Status](https://github.com/esl/segmented_cache/actions/workflows/ci.yml/badge.svg)](https://github.com/esl/segmented_cache/actions/workflows/ci.yml)\n[![codecov](https://codecov.io/gh/esl/segmented_cache/branch/main/graph/badge.svg)](https://codecov.io/gh/esl/segmented_cache)\n[![Hex](http://img.shields.io/hexpm/v/segmented_cache.svg)](https://hex.pm/packages/segmented_cache)\n\n`segmented_cache` is a key-value cache library implemented in rotating segments.","ref":"readme.html"},{"type":"extras","title":"How does it work - Overview","doc":"It uses a coarse-grained strategy, that is, it keeps a set of ETS tables that are periodically rotated, and on rotation, the last table is cleared. It supports diverse eviction strategies, it takes advantage of modern OTP functionality like [`persistent_term`](https://erlang.org/doc/man/persistent_term.html) or [`atomics`](https://erlang.org/doc/man/atomics.html), and it is instrumented with [`telemetry`](https://hexdocs.pm/telemetry/readme.html).\n\nThe cache spawns a process cleaner responsible for periodically cleaning and rotating the tables. It also automatically creates a pg group to sync caches in a cluster. Note that this process will be the owner of the ets tables, so proper care of its supervision must be ensued.","ref":"readme.html#how-does-it-work"},{"type":"extras","title":"Why another caching library - Overview","doc":"Like many things in computing, it depends on the use-case:","ref":"readme.html#why-another-caching-library"},{"type":"extras","title":"Time-to-live - Overview","doc":"Very often, the `ttl` of an entry isn't required to be strictly enforced, and it can instead be allocated in \"buckets\", where the actual `ttl` is allowed to be anything between pre-configured buckets. This simplifies book-keeping, as only the life of the buckets needs to be tracked, instead of a fine-grained per-entry book-keeping.","ref":"readme.html#time-to-live"},{"type":"extras","title":"Extensible values - Overview","doc":"Entries can have any type of values, but upon inserts of the same key, a strategy is often needed to decide the final value: whether we want to always keep the latest, always keep the first, or merge them in some way, a custom-tailored callback can be given to decide. A common pattern can be, for example, to keep maps with metadata, and to use `fun maps:merge/2` as the callback for conflict resolution.\n\n### *Concurrency*\nThis implementation is specially tailored for massively concurrent cache lookups and inserts: many caches in Erlang read from an ets table with `read_concurrency` but serialise inserts through a `gen_server`, which, in high insert bursts, becomes a bottleneck, not considering duplicates inserts that infinitely many workers might request. Other caches serialise read and writes through a `gen_server` that keeps the data in a `map` in its state: this is not only a bottleneck but a source of garbage, as the map is copied on every mutation.\n\nHere instead, the cache uses a –bunch of– ets table with `read_concurrency`, `write_concurrency`, and `decentralized_counters`, to maximise concurrent throughput of all read and writes. It also stores references to the buckets, the index, and the merge strategy, using `persistent_term`, to maximise performance.","ref":"readme.html#extensible-values"},{"type":"extras","title":"Garbage collection and shared memory - Overview","doc":"All operations hove been carefully written following the latest OTP [efficiency guide](https://erlang.org/doc/efficiency_guide/users_guide.html), including maps operations that improve sharing, avoiding unnecessary copying from ETS tables, inline list functions, use `atomics` and `persistent_term` as in [this guide](https://blog.erlang.org/persistent_term/).","ref":"readme.html#garbage-collection-and-shared-memory"},{"type":"extras","title":"Instrumentation - Overview","doc":"These days, all modern services must be instrumented. This cache library helps follow the RED method –Rate, Errors, Duration–: that is, lookup operations raise telemetry events with name `[segmented_cache, Name, request]` with information whether there was a hit or not (`hit := boolean()`), and the time the lookup took, in microseconds (`time := integer()`). With this, we can aggregate the total Rate, and extract the proportion of cache misses, or Errors, while knowing the Duration of the lookups. See the documentation for details.","ref":"readme.html#instrumentation"},{"type":"extras","title":"Configuration - Overview","doc":"* `strategy` can be fifo or lru. Default is `fifo`.\n* `segment_num` is the number of segments for the cache. Default is `3`\n* `ttl` is the live, in minutes, of _each_ segment. Default is `480`, i.e., 8 hours. Can also be a tuple of `{erlang:time_unit(), non_neg_integer()}`\n* `merger_fun` is a callback to resolve conflicts, that takes two conflicting values and returns the final value to be stored. This function takes, in order, the value already present in the table, and the one that is being now inserted. For example, if maps are being used as values, `merger_fun` can be [`fun maps:merge/2`](https://erlang.org/doc/man/maps.html#merge-2).","ref":"readme.html#configuration"},{"type":"extras","title":"Design Notes","doc":"Here are the key points of the design:","ref":"design_notes.html"},{"type":"extras","title":"Cache state - Design Notes","doc":"- The cache records are stored in multiple ETS tables called _segments_, stored in a tuple to improve access times.\n- There's one _current_ table pointed at by an index, stored in an [`atomic`](https://erlang.org/doc/man/atomics.html).\n- The `persistent_term` module is used to store the cache state, which contains the tuple of segments and the atomic reference. It's initialised just once, and it never changes after that.\n- The atomic value is changing, but most importantly: changing in a lock-free manner.\n- Writes are always done at the _current_ table.\n- Reads iterate through all of them in order, starting from the _current_ one.","ref":"design_notes.html#cache-state"},{"type":"extras","title":"TTL implementation - Design Notes","doc":"- Tables are rotated periodically and data from the last table is dropped.\n- `ttl` is not 100% accurate, as a record can be inserted during rotation and therefore live one segment less that expected: we can treat the `ttl` as a warranty that a record will live less than `ttl` but at least `ttl - ttl/N` where `N` is the number of segments (ETS tables).\n- There's a `gen_server` process responsible for the table rotation (see [Cache state notes](#cache-state-notes) for more details).","ref":"design_notes.html#ttl-implementation"},{"type":"extras","title":"LRU implementation - Design Notes","doc":"- On `segmented_cache:is_member/2` and `segmented_cache:get_entry/2` calls, the record is reinserted in the _current_ ETS table.","ref":"design_notes.html#lru-implementation"},{"type":"extras","title":"Distribution - Design Notes","doc":"- In a distributed environment, cache is populated independently on every node.\n- However, we must ensure that on deletion the cache record is removed on all the nodes (and from all the ETS tables, see [LRU implementation notes](#lru-implementation-notes))\n- There's a `gen_server` process responsible for ETS tables rotation (see [TTL implementation notes](#ttl-implementation-notes))\n- The same process is reused to implement asynchronous cache record deletion on other nodes in the cluster.\n- In order to simplify discovery of these `gen_server` processes on other nodes, they all are added into a dedicated `pg` group.","ref":"design_notes.html#distribution"},{"type":"extras","title":"License","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n   Copyright 2021, Erlang Solutions LTD.\n\n   Licensed under the Apache License, Version 2.0 (the \"License\");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an \"AS IS\" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.","ref":"license.html"}],"proglang":"erlang","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.1"}}